"""
Router for the JobSubmission resource.
"""
from typing import Any

from buzz import handle_errors
from fastapi import APIRouter, Depends, HTTPException, Path, Query
from fastapi import Response as FastAPIResponse
from fastapi import status
from fastapi_pagination import Page
from loguru import logger

from jobbergate_api.apps.constants import FileType
from jobbergate_api.apps.dependencies import SecureService, secure_services
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus, slurm_job_state_details
from jobbergate_api.apps.job_submissions.properties_parser import get_job_properties_from_job_script
from jobbergate_api.apps.job_submissions.schemas import (
    ActiveJobSubmission,
    JobSubmissionAgentRejectedRequest,
    JobSubmissionAgentSubmittedRequest,
    JobSubmissionAgentUpdateRequest,
    JobSubmissionCreateRequest,
    JobSubmissionDetailedView,
    JobSubmissionListView,
    JobSubmissionUpdateRequest,
    PendingJobSubmission,
)
from jobbergate_api.apps.permissions import Permissions
from jobbergate_api.apps.schemas import ListParams
from jobbergate_api.apps.services import ServiceError
from jobbergate_api.email_notification import notify_submission_rejected
from jobbergate_api.rabbitmq_notification import publish_status_change

router = APIRouter(prefix="/job-submissions", tags=["Job Submissions"])


@router.post(
    "",
    status_code=status.HTTP_201_CREATED,
    description="Endpoint for job_submission creation",
    response_model=JobSubmissionDetailedView,
)
async def job_submission_create(
    create_request: JobSubmissionCreateRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_email=True)
    ),
):
    """
    Create a new job submission.

    Make a post request to this endpoint with the required values to create a new job submission.
    """
    logger.debug(f"Creating job submissions with {create_request=}")

    client_id = create_request.client_id or secure_services.identity_payload.client_id
    if client_id is None:
        message = "Could not find a client_id in the request body or auth token."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=message,
        )
    create_request.client_id = client_id

    base_job_script = await secure_services.crud.job_script.get(
        create_request.job_script_id, include_files=True
    )

    job_script_files = [f for f in base_job_script.files if f.file_type == FileType.ENTRYPOINT]

    if len(job_script_files) != 1:
        message = "Job script {} has {} entrypoint files, one and only one is required".format(
            create_request.job_script_id, len(job_script_files)
        )
        logger.warning(message)
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=message)

    if create_request.slurm_job_id is None:
        main_file_content = await secure_services.file.job_script.get_file_content(job_script_files[0])

        with handle_errors(
            "Failed to parse execution parameters from the job script",
            raise_exc_class=ServiceError,
            raise_kwargs=dict(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY),
        ):
            execution_parameters = get_job_properties_from_job_script(
                main_file_content.decode(), **create_request.execution_parameters.dict(exclude_unset=True)
            )

        create_request.execution_parameters = execution_parameters
        submission_status = JobSubmissionStatus.CREATED
    else:
        if create_request.execution_parameters.dict(exclude_unset=True):
            message = "Execution parameters are not allowed for on-site job submissions"
            logger.warning(message)
            raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=message)
        submission_status = JobSubmissionStatus.SUBMITTED

    new_job_submission = await secure_services.crud.job_submission.create(
        **create_request.dict(exclude_unset=True),
        owner_email=secure_services.identity_payload.email,
        status=submission_status,
    )
    return new_job_submission


@router.get(
    "/{id}",
    description="Endpoint to get a job_submission",
    response_model=JobSubmissionDetailedView,
)
async def job_submission_get(
    id: int = Path(...),
    secure_services: SecureService = Depends(secure_services(Permissions.JOB_SUBMISSIONS_VIEW, commit=False)),
):
    """Return the job_submission given it's id."""
    logger.debug(f"Getting job submission {id=}")
    return await secure_services.crud.job_submission.get(id)


@router.get(
    "",
    description="Endpoint to list job_submissions",
    response_model=Page[JobSubmissionListView],
)
async def job_submission_get_list(
    list_params: ListParams = Depends(),
    slurm_job_ids: str
    | None = Query(
        None,
        description="Comma-separated list of slurm-job-ids to match active job_submissions",
    ),
    submit_status: JobSubmissionStatus
    | None = Query(
        None,
        description="Limit results to those with matching status",
    ),
    from_job_script_id: int
    | None = Query(
        None,
        description="Filter job-submissions by the job-script-id they were created from.",
    ),
    secure_services: SecureService = Depends(secure_services(Permissions.JOB_SUBMISSIONS_VIEW, commit=False)),
):
    """List job_submissions for the authenticated user."""
    logger.debug("Fetching job submissions")

    list_kwargs = list_params.dict(exclude_unset=True, exclude={"user_only", "include_archived"})

    if list_params.user_only:
        list_kwargs["owner_email"] = secure_services.identity_payload.email
    if submit_status:
        list_kwargs["status"] = submit_status
    if from_job_script_id:
        list_kwargs["job_script_id"] = from_job_script_id
    if slurm_job_ids:
        try:
            list_kwargs["filter_slurm_job_ids"] = [int(i) for i in slurm_job_ids.split(",")]
        except Exception:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Invalid slurm_job_ids param. Must be a comma-separated list of integers",
            )

    return await secure_services.crud.job_submission.paginated_list(
        **list_kwargs,
    )


@router.delete(
    "/{id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete job submission",
)
async def job_submission_delete(
    id: int = Path(..., description="id of the job submission to delete"),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_email=True)
    ),
):
    """Delete job_submission given its id."""
    logger.info(f"Deleting job submission {id=}")
    await secure_services.crud.job_submission.get(
        id, ensure_attributes={"owner_email": secure_services.identity_payload.email}
    )
    await secure_services.crud.job_submission.delete(id)
    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.put(
    "/{id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job_submission given the id",
    response_model=JobSubmissionDetailedView,
)
async def job_submission_update(
    update_params: JobSubmissionUpdateRequest,
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_email=True)
    ),
):
    """Update a job_submission given its id."""
    logger.debug(f"Updating {id=} with {update_params=}")
    await secure_services.crud.job_submission.get(
        id, ensure_attributes={"owner_email": secure_services.identity_payload.email}
    )
    return await secure_services.crud.job_submission.update(id, **update_params.dict(exclude_unset=True))


# The "agent" routes are used for agents to fetch pending job submissions and update their statuses
@router.put(
    "/agent/{id}",
    status_code=status.HTTP_202_ACCEPTED,
    description="Endpoint for an agent to update the status of a job_submission",
    tags=["Agent"],
)
async def job_submission_agent_update(
    update_params: JobSubmissionAgentUpdateRequest,
    id: int = Path(),
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_client_id=True)
    ),
):
    """
    Update a job_submission with slurm_job_state and slurm_job_info.

    Note that if the new slurm_job_state is a termination state, the job submission status will be updated.
    """
    logger.debug(f"Agent is requesting to update {id=}")

    job_submission = await secure_services.crud.job_submission.get(
        id, ensure_attributes={"client_id": secure_services.identity_payload.client_id}
    )

    if job_submission.status != JobSubmissionStatus.SUBMITTED:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Only SUBMITTED jobs may be updated by the agent",
        )

    if job_submission.slurm_job_id != update_params.slurm_job_id:
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail="Update slurm job id does not match the job id on record for job submission {id}",
        )

    logger.info(
        f"Setting slurm job state status to: {update_params.slurm_job_state} "
        f"for job_submission: {id} "
        f"on client_id: {secure_services.identity_payload.client_id}"
    )

    update_dict: dict[str, Any] = dict(
        slurm_job_id=update_params.slurm_job_id,
        slurm_job_state=update_params.slurm_job_state,
        slurm_job_info=update_params.slurm_job_info,
    )

    job_state_details = slurm_job_state_details[update_params.slurm_job_state]
    if job_state_details.is_abort_status:
        update_dict["status"] = JobSubmissionStatus.ABORTED
        update_dict["report_message"] = update_params.slurm_job_state_reason
    elif job_state_details.is_done_status:
        update_dict["status"] = JobSubmissionStatus.DONE

    job_submission = await secure_services.crud.job_submission.update(id, **update_dict)

    if job_submission.status in (
        JobSubmissionStatus.ABORTED,
        JobSubmissionStatus.DONE,
    ):
        await publish_status_change(
            job_submission,
            organization_id=secure_services.identity_payload.organization_id,
        )

    return FastAPIResponse(status_code=status.HTTP_202_ACCEPTED)


@router.post(
    "/agent/submitted",
    description="Endpoint to report that a pending job_submission was submitted",
    tags=["Agent"],
    status_code=status.HTTP_202_ACCEPTED,
)
async def job_submissions_agent_submitted(
    submitted_request: JobSubmissionAgentSubmittedRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_client_id=True)
    ),
):
    """Update a job_submission to indicate that it was submitted to Slurm."""
    logger.debug("Agent is reporting that a pending job has been submitted")

    job_submission = await secure_services.crud.job_submission.get(
        submitted_request.id, ensure_attributes={"client_id": secure_services.identity_payload.client_id}
    )
    if job_submission.status != JobSubmissionStatus.CREATED:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Only CREATED Job Submissions can be marked as SUBMITTED",
        )

    logger.info(f"Marking job_submissions {submitted_request.id} as SUBMITTED")

    await secure_services.crud.job_submission.update(
        submitted_request.id,
        slurm_job_id=submitted_request.slurm_job_id,
        status=JobSubmissionStatus.SUBMITTED,
    )
    return FastAPIResponse(status_code=status.HTTP_202_ACCEPTED)


@router.post(
    "/agent/rejected",
    description="Endpoint to report that a pending job_submission was rejected by Slurm",
    tags=["Agent"],
    status_code=status.HTTP_202_ACCEPTED,
)
async def job_submissions_agent_rejected(
    rejected_request: JobSubmissionAgentRejectedRequest,
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_EDIT, ensure_client_id=True)
    ),
):
    """Update a job_submission to indicate that it was rejected by Slurm."""
    logger.debug("Agent is reporting that a pending job has been rejected")

    job_submission = await secure_services.crud.job_submission.get(
        rejected_request.id, ensure_attributes={"client_id": secure_services.identity_payload.client_id}
    )
    if job_submission.status != JobSubmissionStatus.CREATED:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Only CREATED Job Submissions can be marked as REJECTED",
        )

    logger.info(f"Marking job_submissions {rejected_request.id} as REJECTED")

    await secure_services.crud.job_submission.update(
        rejected_request.id,
        status=JobSubmissionStatus.REJECTED,
        report_message=rejected_request.report_message,
    )
    notify_submission_rejected(
        rejected_request.id,
        rejected_request.report_message,
        job_submission.owner_email,
    )

    await publish_status_change(
        job_submission,
        organization_id=secure_services.identity_payload.organization_id,
    )

    return FastAPIResponse(status_code=status.HTTP_202_ACCEPTED)


@router.get(
    "/agent/pending",
    description="Endpoint to list pending job_submissions for the requesting client",
    response_model=Page[PendingJobSubmission],
    response_model_exclude_none=True,
    tags=["Agent"],
)
async def job_submissions_agent_pending(
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_VIEW, commit=False, ensure_client_id=True)
    ),
):
    """Get a list of pending job submissions for the cluster-agent."""
    logger.debug("Agent is requesting a list of pending job submissions")

    client_id = secure_services.identity_payload.client_id

    logger.info(f"Fetching newly created job_submissions for {client_id=}")

    return await secure_services.crud.job_submission.paginated_list(
        status=JobSubmissionStatus.CREATED,
        client_id=client_id,
        include_files=True,
        include_parent=True,
    )


@router.get(
    "/agent/active",
    description="Endpoint to list active job_submissions for the requesting client",
    response_model=Page[ActiveJobSubmission],
    tags=["Agent"],
)
async def job_submissions_agent_active(
    secure_services: SecureService = Depends(
        secure_services(Permissions.JOB_SUBMISSIONS_VIEW, commit=False, ensure_client_id=True)
    ),
):
    """Get a list of active job submissions for the cluster-agent."""
    logger.debug("Agent is requesting a list of active job submissions")

    client_id = secure_services.identity_payload.client_id

    logger.info(f"Fetching active job_submissions for {client_id=}")

    pages = await secure_services.crud.job_submission.paginated_list(
        status=JobSubmissionStatus.SUBMITTED,
        client_id=client_id,
    )
    return pages



"""
Database model for the JobSubmission resource.
"""

from __future__ import annotations

from typing import Any

from sqlalchemy import Enum, ForeignKey, Integer, String, text
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import Mapped, mapped_column, relationship, selectinload
from sqlalchemy.sql.expression import Select

from jobbergate_api.apps.job_scripts.models import JobScript as JobScriptModel
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus, SlurmJobState
from jobbergate_api.apps.models import Base, CrudMixin
from jobbergate_api.safe_types import JobScript


class JobSubmission(CrudMixin, Base):
    """
    Job submission table definition.

    Notice all relationships are lazy="raise" to prevent n+1 implicit queries.
    This means that the relationships must be explicitly eager loaded using
    helper functions in the class.

    Attributes:
        job_script_id: Id number of the job scrip this submissions is based on.
        execution_directory: The directory where the job is executed.
        slurm_job_id: The id of the job in the slurm queue.
        slurm_job_state: The Slurm Job state as reported by the agent
        slurm_job_info: Detailed information about the  Slurm Job as reported by the agent
        client_id: The id of the custer this submission runs on.
        status: The status of the job submission.
        report_message: The message returned by the job.
        execution_parameters: The properties of the job.

    See Mixin class definitions for other columns
    """

    job_script_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("job_scripts.id", ondelete="SET NULL"),
        nullable=True,
    )
    execution_directory: Mapped[str] = mapped_column(String, default=None, nullable=True)

    slurm_job_id: Mapped[int] = mapped_column(Integer, default=None, nullable=True)
    slurm_job_state: Mapped[SlurmJobState] = mapped_column(
        Enum(SlurmJobState, native_enum=False),
        nullable=True,
    )
    slurm_job_info: Mapped[str] = mapped_column(String, default=None, nullable=True)

    client_id: Mapped[str] = mapped_column(String, nullable=False, index=True)
    status: Mapped[JobSubmissionStatus] = mapped_column(
        Enum(JobSubmissionStatus, native_enum=False),
        nullable=False,
        index=True,
    )
    report_message: Mapped[str] = mapped_column(String, nullable=True)

    execution_parameters: Mapped[dict[str, Any]] = mapped_column(
        "template_vars",
        JSONB,
        nullable=False,
        default=text("'{}'::jsonb"),
        server_default=text("'{}'::jsonb"),
    )

    job_script: Mapped[JobScript] = relationship(
        "JobScript",
        back_populates="submissions",
        lazy="raise",
    )

    @classmethod
    def searchable_fields(cls):
        """
        Add client_id as a searchable field.
        """
        return {cls.client_id, *super().searchable_fields()}

    @classmethod
    def sortable_fields(cls):
        """
        Add additional sortable fields.
        """
        return {
            cls.job_script_id,
            cls.slurm_job_id,
            cls.client_id,
            cls.status,
            *super().sortable_fields(),
        }

    @classmethod
    def include_files(cls, query: Select) -> Select:
        """
        Include custom options on a query to eager load files.
        """
        return query.options(selectinload(cls.job_script).options(selectinload(JobScriptModel.files)))

    @classmethod
    def include_parent(cls, query: Select) -> Select:
        """
        Include custom options on a query to eager load parent data.
        """
        return query.options(selectinload(cls.job_script))



"""
Parser for Slurm REST API parameters from SBATCH parameters at the job script file.
"""
from argparse import ArgumentParser
from dataclasses import dataclass, field
from itertools import chain
from shlex import split
from typing import Any, Dict, List, Union

from bidict import bidict
from loguru import logger

from jobbergate_api.apps.job_submissions.schemas import JobProperties

_IDENTIFICATION_FLAG = "#SBATCH"
_INLINE_COMMENT_MARK = "#"


def _flagged_line(line: str) -> bool:
    """
    Identify if a provided line starts with the identification flag.
    """
    return line.startswith(_IDENTIFICATION_FLAG)


def _clean_line(line: str) -> str:
    """
    Clean the provided line.

    It includes removing the identification flag at the beginning of the line,
    then remove the inline comment mark and anything after it, and finally strip white spaces at both sides.
    """
    return line.lstrip(_IDENTIFICATION_FLAG).split(_INLINE_COMMENT_MARK)[0].strip()


def _clean_jobscript(jobscript: str) -> List[str]:
    """
    Transform a job script string.

    It is done by filtering only the lines that start with
    the identification flag and mapping a cleaning procedure to them in order
    to remove the identification flag, remove inline comments, and strip extra
    white spaces. Finally, split each pair of parameter/value and chain them
    in a single list.
    """
    jobscript_filtered = filter(_flagged_line, jobscript.splitlines())
    jobscript_cleaned = map(_clean_line, jobscript_filtered)
    jobscript_splitted = map(split, jobscript_cleaned)
    return list(chain.from_iterable(jobscript_splitted))


@dataclass(frozen=True)
class SbatchToSlurm:
    """
    Store the information for each parameter, including its name at Slurm API and SBATCH.

    Besides that, any extra argument this parameter needs when added to
    the parser. This information is used to build the jobscript/SBATCH parser
    and the two-way mapping between Slurm API and SBATCH names.
    """

    slurmrestd_var_name: str
    sbatch: str
    sbatch_short: str = ""
    argparser_param: dict = field(default_factory=dict)


sbatch_to_slurm_mapping = [
    SbatchToSlurm("account", "--account", "-A"),
    SbatchToSlurm("account_gather_frequency", "--acctg-freq"),
    SbatchToSlurm("array", "--array", "-a"),
    SbatchToSlurm("batch_features", "--batch"),
    SbatchToSlurm("burst_buffer", "--bb"),
    SbatchToSlurm("", "--bbf"),
    SbatchToSlurm("begin_time", "--begin", "-b"),
    SbatchToSlurm("current_working_directory", "--chdir", "-D"),
    SbatchToSlurm("cluster_constraints", "--cluster-constraint"),
    SbatchToSlurm("", "--clusters", "-M"),
    SbatchToSlurm("comment", "--comment"),
    SbatchToSlurm("constraints", "--constraint", "-C"),
    SbatchToSlurm("container", "--container"),
    SbatchToSlurm("", "--contiguous", "", dict(action="store_const", const=True)),
    SbatchToSlurm("core_specification", "--core-spec", "-S", dict(type=int)),
    SbatchToSlurm("cores_per_socket", "--cores-per-socket", "", dict(type=int)),
    SbatchToSlurm("cpu_binding", "--cpu-bind"),
    SbatchToSlurm("cpu_frequency", "--cpu-freq"),
    SbatchToSlurm("cpus_per_gpu", "--cpus-per-gpu"),
    SbatchToSlurm("cpus_per_task", "--cpus-per-task", "-c", dict(type=int)),
    SbatchToSlurm("deadline", "--deadline"),
    SbatchToSlurm("delay_boot", "--delay-boot", "", dict(type=int)),
    SbatchToSlurm("dependency", "--dependency", "-d"),
    SbatchToSlurm("distribution", "--distribution", "-m"),
    SbatchToSlurm("standard_error", "--error", "-e"),
    SbatchToSlurm("", "--exclude", "-x"),
    SbatchToSlurm(
        "exclusive",
        "--exclusive",
        "",
        dict(
            type=str,
            choices={"user", "mcs", "exclusive", "oversubscribe"},
            nargs="?",
            const="exclusive",
        ),
    ),
    SbatchToSlurm("", "--export"),
    SbatchToSlurm("", "--export-file"),
    SbatchToSlurm("", "--extra-node-info", "-B"),
    SbatchToSlurm("get_user_environment", "--get-user-env", "", dict(type=int)),
    SbatchToSlurm("", "--gid"),
    SbatchToSlurm("gpu_binding", "--gpu-bind"),
    SbatchToSlurm("gpu_frequency", "--gpu-freq"),
    SbatchToSlurm("gpus", "--gpus", "-G"),
    SbatchToSlurm("gpus_per_node", "--gpus-per-node"),
    SbatchToSlurm("gpus_per_socket", "--gpus-per-socket"),
    SbatchToSlurm("gpus_per_task", "--gpus-per-task"),
    SbatchToSlurm("gres", "--gres"),
    SbatchToSlurm("gres_flags", "--gres-flags"),
    SbatchToSlurm("", "--hint"),
    SbatchToSlurm("hold", "--hold", "-H", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--ignore-pbs", "", dict(action="store_const", const=True)),
    SbatchToSlurm("standard_input", "--input", "-i"),
    SbatchToSlurm("name", "--job-name", "-J"),
    SbatchToSlurm(
        "kill_on_invalid_dependency", "--kill-on-invalid-dep", "", dict(type=bool, nargs="?", const=True)
    ),
    SbatchToSlurm("licenses", "--licenses", "-L"),
    SbatchToSlurm("mail_type", "--mail-type"),
    SbatchToSlurm("mail_user", "--mail-user"),
    SbatchToSlurm("mcs_label", "--mcs-label"),
    SbatchToSlurm("memory_per_node", "--mem"),
    SbatchToSlurm("memory_binding", "--mem-bind"),
    SbatchToSlurm("memory_per_cpu", "--mem-per-cpu"),
    SbatchToSlurm("memory_per_gpu", "--mem-per-gpu"),
    SbatchToSlurm("minimum_cpus_per_node", "--mincpus", "", dict(type=int)),
    SbatchToSlurm("", "--network"),
    SbatchToSlurm("nice", "--nice"),
    SbatchToSlurm("no_kill", "--no-kill", "-k", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--no-requeue", "", dict(action="store_false", dest="requeue", default=None)),
    SbatchToSlurm("", "--nodefile", "-F"),
    SbatchToSlurm("", "--nodelist", "-w"),
    SbatchToSlurm("nodes", "--nodes", "-N"),
    SbatchToSlurm("tasks", "--ntasks", "-n", dict(type=int)),
    SbatchToSlurm("tasks_per_core", "--ntasks-per-core", "", dict(type=int)),
    SbatchToSlurm("", "--ntasks-per-gpu"),
    SbatchToSlurm("tasks_per_node", "--ntasks-per-node", "", dict(type=int)),
    SbatchToSlurm("tasks_per_socket", "--ntasks-per-socket", "", dict(type=int)),
    SbatchToSlurm("open_mode", "--open-mode"),
    SbatchToSlurm("standard_output", "--output", "-o"),
    SbatchToSlurm("", "--overcommit", "-O", dict(action="store_const", const=True)),
    SbatchToSlurm(
        "",
        "--oversubscribe",
        "-s",
        dict(action="store_const", const="oversubscribe", dest="exclusive"),
    ),
    SbatchToSlurm("", "--parsable", "", dict(action="store_const", const=True)),
    SbatchToSlurm("partition", "--partition", "-p"),
    SbatchToSlurm("", "--power"),
    SbatchToSlurm("priority", "--priority"),
    SbatchToSlurm("", "--profile"),
    SbatchToSlurm("", "--propagate"),
    SbatchToSlurm("qos", "--qos", "-q"),
    SbatchToSlurm("", "--quiet", "-Q", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--reboot", "", dict(action="store_const", const=True)),
    SbatchToSlurm("requeue", "--requeue", "", dict(action="store_true", default=None)),
    SbatchToSlurm("reservation", "--reservation"),
    SbatchToSlurm("signal", "--signal"),
    SbatchToSlurm("sockets_per_node", "--sockets-per-node", "", dict(type=int)),
    SbatchToSlurm("spread_job", "--spread-job", "", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--switches"),
    SbatchToSlurm("", "--test-only", "", dict(action="store_const", const=True)),
    SbatchToSlurm("thread_specification", "--thread-spec", "", dict(type=int)),
    SbatchToSlurm("threads_per_core", "--threads-per-core", "", dict(type=int)),
    SbatchToSlurm("time_limit", "--time", "-t"),
    SbatchToSlurm("time_minimum", "--time-min"),
    SbatchToSlurm("", "--tmp"),
    SbatchToSlurm("", "--uid"),
    SbatchToSlurm("", "--usage", "", dict(action="store_const", const=True)),
    SbatchToSlurm("minimum_nodes", "--use-min-nodes", "", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--verbose", "-v", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--version", "-V", dict(action="store_const", const=True)),
    SbatchToSlurm("", "--wait", "-W", dict(action="store_const", const=True)),
    SbatchToSlurm("wait_all_nodes", "--wait-all-nodes", "", dict(type=int)),
    SbatchToSlurm("wckey", "--wckey"),
    SbatchToSlurm("", "--wrap"),
]


class ArgumentParserCustomExit(ArgumentParser):
    """
    Custom implementation of the built-in class for argument parsing.

    The sys.exit triggered by the original code is replaced by a ValueError,
    besides some friendly logging messages.
    """

    def exit(self, status=0, message=None):
        """
        Raise ValueError when parsing invalid parameters or if the type of their values is not correct.
        """
        log_message = f"Argparse exit status {status}: {message}"
        if status:
            logger.error(log_message)
        else:
            logger.info(log_message)
        raise ValueError(message)


def build_parser() -> ArgumentParser:
    """
    Build an ArgumentParser to handle all SBATCH parameters declared at sbatch_to_slurm.
    """
    parser = ArgumentParserCustomExit()
    for item in sbatch_to_slurm_mapping:
        args = (i for i in (item.sbatch_short, item.sbatch) if i)
        parser.add_argument(*args, **item.argparser_param)

    return parser


def build_mapping_sbatch_to_slurm() -> bidict:
    """
    Create a mapper to translate in both ways between the names expected by Slurm REST API and SBATCH.
    """
    mapping: bidict = bidict()

    for item in sbatch_to_slurm_mapping:
        if item.slurmrestd_var_name:
            sbatch_name = item.sbatch.lstrip("-").replace("-", "_")
            mapping[sbatch_name] = item.slurmrestd_var_name

    return mapping


def jobscript_to_dict(jobscript: str) -> Dict[str, Union[str, bool]]:
    """
    Extract the SBATCH params from a given job script.

    It returns them in a dictionary for mapping the parameter names to their values.

    Raise ValueError if any of the parameters are unknown to the parser.
    """
    parsed_args, unknown_arg = parser.parse_known_args(args=_clean_jobscript(jobscript))

    if unknown_arg:
        raise ValueError("Unrecognized SBATCH arguments: {}".format(" ".join(unknown_arg)))

    sbatch_params = {key: value for key, value in vars(parsed_args).items() if value is not None}

    logger.debug(f"SBATCH params parsed from job script: {sbatch_params}")

    return sbatch_params


def convert_sbatch_to_slurm_api(input: Dict[str, Any]) -> Dict[str, Any]:
    """
    Take a dictionary containing key-value pairing of SBATCH parameter name space to Slurm API namespace.

    Notice the values should not be affected.

    Raise KeyError if any of the keys are unknown to the mapper.
    """
    mapped = {}
    unknown_keys = []

    for sbatch_name, value in input.items():
        try:
            slurm_name = mapping_sbatch_to_slurm[sbatch_name]
            mapped[slurm_name] = value
        except KeyError:
            unknown_keys.append(sbatch_name)

    if unknown_keys:
        error_message = "Impossible to convert from SBATCH to Slurm REST API: {}"
        raise KeyError(error_message.format(", ".join(unknown_keys)))

    logger.debug(f"Slurm API params mapped from SBATCH: {mapped}")

    return mapped


def get_job_parameters(jobscript: str) -> Dict[str, Any]:
    """
    Parse all SBATCH parameters from a job script, map their names to Slurm API parameters.

    They are returned as a key-value pairing dictionary.
    """
    return convert_sbatch_to_slurm_api(jobscript_to_dict(jobscript))


def get_job_properties_from_job_script(main_file_content: str, **kwargs) -> JobProperties:
    """
    Get the job properties for Slurm REST API from a job script file.

    Extra keyword arguments can be used to overwrite any parameter from the
    job script, like name or current_working_directory.
    """
    slurm_parameters = get_job_parameters(main_file_content)
    merged_parameters = {**slurm_parameters, **kwargs}
    return JobProperties.parse_obj(merged_parameters)


parser = build_parser()
mapping_sbatch_to_slurm = build_mapping_sbatch_to_slurm()
